#
#  gpsman --- GPS Manager: a manager for GPS receiver data
#
#  Copyright (c) 2000 Miguel Filgueiras (mig@ncc.up.pt) / Universidade do Porto
#
#    This program is free software; you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation; either version 2 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.
#
#  File: compute.tcl
#  Last change:  8 November 2000
#

## Some formulae kindly supplied by
#     Luisa Bastos, Universidade do Porto
#     Gil Goncalves, Universidade de Coimbra
##
#
## Algorithms and formulae adopted from:
#     - GKK grid conversion by Andreas Lange (andreas.lange@rhein-main.de)
#       in GPStrans 0.36
##

### positions and coordinates

array set GRIDZN {
    BNG  {([A-H]|[J-Z])([A-H]|[J-Z])}
    CMP  {}
    GKK  {[0-5]}
    ITM  {[A-H]|[J-Z]}
    KKJY {27E}
    KKJP {[1-4]}
    SEG  {}
}

set GRIDS [lsort [array names GRIDZN]]
set PFORMATS [concat [list $TXT(DMS) $TXT(DMM) $TXT(DDD) $TXT(UTM/UPS)] $GRIDS]

foreach g $GRIDS {
    set TXT($g) $g ; set INVTXT($g) $g
}

proc PosType {pformt} {

    switch $pformt {
	DMS -
	DMM -
	DDD     { return latlong }
	UTM/UPS { return utm }
	default { return grid }
    }
}

proc Coord {pformt coord negh} {
    # convert coordinate $coord in format $pformt, with negative heading $negh,
    #  to signed degrees
    # $pformt in {DMS, DMM, DDD}

    set coord [string trim $coord]
    set sign 1
    set h [string index "$coord" 0]
    if { ! [regexp {[0-9]} $h] } {
	if { "$h" == "$negh" || "$h" == "-" } { set sign -1 }
	set coord [string range "$coord" 1 end]
    }
    switch $pformt {
	DMS {
	    scan "$coord" "%d %d %f" d m s
	    return [expr $sign*($d+$m/60.0+$s/3600.0)]
	}
	DMM {
	    scan "$coord" "%d %f" d m
	    return [expr $sign*($d+$m/60.0)]
	}
	DDD {
	    return [expr $sign*$coord]
	}
    }
}

proc CreatePos {latd longd pformt type datum} {
    # create position representation under format $pformt, type,
    #  from lat/long in degrees, and datum
    # a position representation is a list whose first two elements
    #  are lat and long in degrees, and whose rest is dependent on format:
    #  $type==latlong: lat and long in external format
    #  $type==utm: zones east and north and x y coordinates
    #  $type==grid: zone name and x y coordinates

    switch $type {
	latlong {
	    if { $latd < 0 } {
		set la [expr -$latd] ; set hlat S
	    } else {
		set la $latd ; set hlat N
	    }
	    if { $longd < 0 } {
		set lo [expr -$longd] ; set hlng W
	    } else {
		set lo $longd ; set hlng E
	    }
	    return [list $latd $longd "$hlat[ExtDegrees $pformt $la]" \
		         "$hlng[ExtDegrees $pformt $lo]"]
	}
	utm {
	    set p [DegreesToUTM $latd $longd $datum]
	    return [linsert $p 0 $latd $longd]
	}
	grid {
	    set p [DegreesTo$pformt $latd $longd $datum]
	    return [linsert $p 0 $latd $longd]
	}
    }
}

proc ExtDegrees {pformt degs} {
    # from signed degrees $degs to external representation of format $pformt
    # $pformt in {DMS, DMM, DDD}

    switch $pformt {
	DMS {
	      if { $degs < 0 } {
		  set degs [expr -$degs] ; set sign -1
	      } else { set sign 1 }
	      set d [expr int($degs)]
	      set degs [expr ($degs-$d)*60]
	      set d [expr $sign*$d]
	      set m [expr int($degs)]
	      set s [expr ($degs-$m)*60]
	      if { $s > 59.95 } { set s 0 ; incr m }
	      if { $m > 59 } { set m 0 ; incr d }
	      return [format "%d %02d %04.1f" $d $m $s]
	}
	DMM {
	      if { $degs < 0 } {
		  set degs [expr -$degs] ; set sign -1
	      } else { set sign 1 }
	      set d [expr int($degs)]
	      set degs [expr ($degs-$d)*60]
	      set d [expr $sign*$d]
	      if { $degs > 59.995 } { set degs 0 ; incr d }
	      return [format "%d %06.3f" $d $degs]
	}
	DDD {
	      return [format "%.5f" $degs]
	}
    }
}

proc ChangeCoordSign {pos} {

    if { [string compare $pos N] } {
	return W
    }
    return S
}

### distances and geographic bearings

proc ComputeDist {p1 p2 datum} {
    # distance between positions $p1 and $p2 with same datum
            # formulae kindly supplied by Luisa Bastos, Universidade do Porto
            #   and Gil Goncalves, Universidade de Coimbra

    set lad1 [lindex $p1 0] ; set lod1 [lindex $p1 1]
    set lad2 [lindex $p2 0] ; set lod2 [lindex $p2 1]
    if { $lad1==$lad2 && $lod1==$lod2 } { return 0 }
    set la1 [expr $lad1*0.01745329251994329576]
    set lo1 [expr $lod1*0.01745329251994329576]
    set la2 [expr $lad2*0.01745329251994329576]
    set lo2 [expr $lod2*0.01745329251994329576]
    return [expr 1e-3*[lindex [EllipsdData $datum] 0]* \
	         acos(cos($lo1-$lo2)*cos($la1)*cos($la2)+sin($la1)*sin($la2))]
}

proc ComputeDistBear {p1 p2 datum} {
    # distance between and bearing from positions $p1 and $p2 with same datum
            # formulae kindly supplied by Luisa Bastos, Universidade do Porto
            #   and Gil Goncalves, Universidade de Coimbra

    set lad1 [lindex $p1 0] ; set lod1 [lindex $p1 1]
    set lad2 [lindex $p2 0] ; set lod2 [lindex $p2 1]
    if { $lad1==$lad2 && $lod1==$lod2 } { return "0 0" }
    set la1 [expr $lad1*0.01745329251994329576]
    set lo1 [expr $lod1*0.01745329251994329576]
    set la2 [expr $lad2*0.01745329251994329576]
    set lo2 [expr $lod2*0.01745329251994329576]
    set dt [EllipsdData $datum]
    set a [lindex $dt 0] ; set f [lindex $dt 1]
           # distance
    set d [expr 1e-3*$a*acos(cos($lo1-$lo2)*cos($la1)*cos($la2)+ \
	                 sin($la1)*sin($la2))]
           # bearing
    set da [expr $la2-$la1] ; set do [expr $lo2-$lo1]
    if { [expr abs($da)] < 1e-20 } {
 	if { [expr abs($do)] < 1e-20 } {
 	    set b 0
 	} elseif { $do < 0 } {
 	    set b 270
 	} else { set b 90 }
    } elseif { [expr abs($do)] < 1e-20 } {
 	if { $da < 0 } {
 	    set b 180
 	} else { set b 0 }
    } else {
 	set b [expr round(atan(sin($do)/ \
 		               (tan($la2)*cos($la1)-sin($la1)*cos($do))) \
 			  *57.29577951308232087684)]
	if { $b < 0 } {
	    if { $do < 0 } { incr b 360 } else { incr b 180 }
	} elseif { $do < 0 } { incr b 180 }
    }
    return [list $d $b]
}

proc SetDatumData {datum} {
    # this proc sets datum parameters as global variables for repeated use
    #  in conversions; see procs ComputeDistFD, ComputeDistBearFD
    global DatumA DatumF FDDatum

    if { [string compare $FDDatum $datum] } {
	set dt [EllipsdData $datum]
	set DatumA [lindex $dt 0] ; set DatumF [lindex $dt 1]
	set FDDatum $datum
    }
    return
}

proc ComputeDistFD {p1 p2} {
    # compute distance between positions $p1 and $p2 assuming datum
    #  parameters where set by calling SetDatumData
    global DatumA
            # formulae kindly supplied by Luisa Bastos, Universidade do Porto
            #   and Gil Goncalves, Universidade de Coimbra

    set lad1 [lindex $p1 0] ; set lod1 [lindex $p1 1]
    set lad2 [lindex $p2 0] ; set lod2 [lindex $p2 1]
    if { $lad1==$lad2 && $lod1==$lod2 } { return 0 }
    set la1 [expr $lad1*0.01745329251994329576]
    set lo1 [expr $lod1*0.01745329251994329576]
    set la2 [expr $lad2*0.01745329251994329576]
    set lo2 [expr $lod2*0.01745329251994329576]
    return [expr 1e-3*$DatumA*acos(cos($lo1-$lo2)*cos($la1)*cos($la2)+ \
	                      sin($la1)*sin($la2))]
}

proc ComputeDistBearFD {p1 p2} {
    # compute distance between and bearing from positions $p1 and $p2
    #  assuming datum parameters where set by calling SetDatumData
    global DatumA DatumF
            # formulae kindly supplied by Luisa Bastos, Universidade do Porto
            #   and Gil Goncalves, Universidade de Coimbra

    set lad1 [lindex $p1 0] ; set lod1 [lindex $p1 1]
    set lad2 [lindex $p2 0] ; set lod2 [lindex $p2 1]
    if { $lad1==$lad2 && $lod1==$lod2 } { return "0 0" }
    set la1 [expr $lad1*0.01745329251994329576]
    set lo1 [expr $lod1*0.01745329251994329576]
    set la2 [expr $lad2*0.01745329251994329576]
    set lo2 [expr $lod2*0.01745329251994329576]
           # distance
    set d [expr 1e-3*$DatumA*acos(cos($lo1-$lo2)*cos($la1)*cos($la2)+ \
	                 sin($la1)*sin($la2))]
           # bearing
    set da [expr $la2-$la1] ; set do [expr $lo2-$lo1]
    if { [expr abs($da)] < 1e-20 } {
 	if { [expr abs($do)] < 1e-20 } {
 	    set b 0
 	} elseif { $do < 0 } {
 	    set b 270
 	} else { set b 90 }
    } elseif { [expr abs($do)] < 1e-20 } {
 	if { $da < 0 } {
 	    set b 180
 	} else { set b 0 }
    } else {
 	set b [expr round(atan(sin($do)/ \
 		               (tan($la2)*cos($la1)-sin($la1)*cos($do))) \
 			  *57.29577951308232087684)]
	if { $b < 0 } {
	    if { $do < 0 } { incr b 360 } else { incr b 180 }
	} elseif { $do < 0 } { incr b 180 }
    }
    return [list $d $b]
}

proc CompWPDistBear {wp1 wp2} {
    # distance between and bearing from WPs with names $wp1, $wp2
    global WPPFrmt WPPosn WPDatum

    set i1 [IndexNamed WP "$wp1"]
    set i2 [IndexNamed WP "$wp2"]
    if { $i1<0 || $i2<0 } { return "--- ---" }
    if { $i1 == $i2 } { return "0 0" }
    return [CompDistBearDatums $WPPosn($i1) $WPDatum($i1) \
	                          $WPPosn($i2) $WPDatum($i2)]
}

proc CompDistBearDatums {p1 d1 p2 d2} {
    # distance between and bearing from positions $p1 and $p2
    #  with datums $d1 and $d2
    if { [string compare $d1 $d2] } {
	set p2 [ConvertDatum [lindex $p2 0] [lindex $p2 1] $d2 $d1 DDD]
    }
    return [ComputeDistBear $p1 $p2 $d1]
}

###

proc CompUTMOnZone {lat long longz} {
    # UTM/UPS position representation of position with lat/long signed degrees
    #  on zone containing meridian with longitude $longz (signed degrees)
    global UTMlat0 UTMk0 Datum

    if { $lat>=-80 && $lat<=84 } {
	set long0 [expr 6*int(floor($longz/6.0))+3]
	set cs [ConvToTM $lat $long $UTMlat0 $long0 $UTMk0 $Datum]
	set x [expr int(5e5+[lindex $cs 0])] ; set y [lindex $cs 1]
	if { $lat < 0 } { set y [expr int(1e7+$y)] }
	return [list $x $y]
    }
    return [ConvToUPS $lat $long $Datum]
}

proc Long0UTMZone {ze} {
    # compute longitude of central meridian of UTM zones with given
    #  easting number (assumed different from 0)

    return [expr -183+6*$ze]
}

### CMP, GKK and basic KKJ grid coordinates

proc DegreesToCMP {lat long datum} {
    # convert lat/long in signed degrees to Portuguese Military grid coords

    if { $long < -12 || $long > -4 || $lat < 30 || $lat > 50 } {
	return "-- 0 0"
    }
    set cs [ConvToTM $lat $long 39.66666666666667 -8.13190611111111 1.0 $datum]
    if { [set x [expr round([lindex $cs 0]+2e5)]] < 0 || \
	    [set y [expr round([lindex $cs 1]+3e5)]] < 0 } {
	return "-- 0 0"
    }
    return [list "" $x $y]
}

proc CMPToDegrees {zone x y datum} {
    # convert Portuguese Military grid coords to lat/long in signed degrees

    if { "$zone" != "" || $x < 0 || $y < 0 } { return 0 }
    set x [expr $x-2e5] ; set y [expr $y-3e5]
    return [ConvFromTM $x $y 39.66666666666667 -8.13190611111111 1.0 $datum]
}

set GKKZONE 012345

proc DegreesToGKK {lat long datum} {
    # convert from lat/long in signed degrees to German Krueger grid coords
    global GKKZONE

    if { $long < -1.5 || $long > 16.5 || $lat < 0 } {
	return "-- 0 0"
    }
    set z [expr int(($long+1.5)/3)]
    set long0 [expr 3.0*$z]
    set cs [ConvToTM $lat $long 0 $long0 1.0 $datum]
    if { [set x [expr round([lindex $cs 0]+5e5+1e6*$z)]] < 0 || \
	    [set y [expr round([lindex $cs 1])]] < 0 } {
	return "-- 0 0"
    }
    set zc [string index $GKKZONE $z]
    return [list $zc $x $y]
}

proc GKKToDegrees {zone x y datum} {
    # convert from German Krueger grid coords to lat/long in signed degrees
    global GKKZONE

    if { $x < 0 || $y < 0 || $x > 6e6 || $y > 1e7 || \
	    [set z [string first $zone $GKKZONE]] == -1 } {
	return 0
    }
    set long0 [expr 3.0*$z]
    set x [expr $x-5e5-1e6*$z]
    return [ConvFromTM $x $y 0 $long0 1.0 $datum]
}

proc DegreesToKKJP {lat long datum} {
    # convert from lat/long in signed degrees to basic Finnish grid coords

    if { $long < 19.5 || $long > 31.5 || $lat < 0 } {
	return "-- 0 0"
    }
    set z [expr int(($long-16.5)/3)]
    set long0 [expr 3.0*$z+18]
    set cs [ConvToTM $lat $long 0 $long0 1.0 $datum]
    if { [set x [expr round([lindex $cs 0]+5e5+1e6*$z)]] < 0 || \
	    [set y [expr round([lindex $cs 1])]] < 0 } {
	return "-- 0 0"
    }
    return [list $z $x $y]
}

proc KKJPToDegrees {zone x y datum} {
    # convert from basic Finnish grid coords to lat/long in signed degrees

    if { $x < 0 || $y < 0 || $x > 5e6 || $y > 1e7 || \
	    ! [regexp {[1-4]} $zone] } {
	return 0
    }
    set long0 [expr 3.0*$zone+18]
    set x [expr $x-5e5-1e6*$zone]
    return [ConvFromTM $x $y 0 $long0 1.0 $datum]
}

### datums

proc ChangeTPsDatum {tps datum1 datum2} {
    # convert position of TR points on list $tps from $datum1 to $datum2
    # changes only on the first 4 elements (position) of each TP representation

    if {! [string compare $datum1 $datum2] } { return $tps }
    set l ""
    foreach tp $tps {
	set np [ConvertDatum [lindex $tp 0] [lindex $tp 1] \
		             $datum1 $datum2 DMS]
	set ntp [concat $np [lrange $tp 4 end]]
	lappend l $ntp
    }
    return $l
}

### dates

proc Today {dformt} {
    # build representation of current date under format $dformt

    set dt [clock format [clock seconds] -format "%Y %m %d %H %M %S"]
    scan $dt %d%0d%0d%0d%0d%0d y m d h mn s
    return [FormatDate $dformt $y $m $d $h $mn $s]
}

proc FormatDate {dformt y m d h mn s} {
    # build representation of date under format $dformt
    #  $dformt in {YYYYMMDD, MMDDYYYY, DDMMMYYYY}
    # result assumed elsewhere to be a list whose head is the information on
    #  year, month and day
    global MONTHNAMES

    set h [format "%02d:%02d:%02d" $h $mn $s]
    switch $dformt {
	YYYYMMDD { return [format "%4d.%02d.%02d %s" $y $m $d $h] }
	MMDDYYYY { return [format "%02d/%02d/%4d %s" $m $d $y $h] }
	DDMMMYYYY {
	    incr m -1 ; set m [lindex $MONTHNAMES $m]
	    return [format "%02d-%s-%4d %s" $d $m $y $h]
	}
    }
}

proc FormatTime {secs} {
    # build represention of seconds in hours:minutes:seconds format

    set s [expr $secs%60] ; set x [expr ($secs-$s)/60]
    set mn [expr $x%60] ; set h [expr ($x-$mn)/60]
    if { $h > 0 } { return [format "%2d:%02d:%02d" $h $mn $s] }
    if { $mn > 0 } { return [format "%5d:%02d" $mn $s] }
    return [format "%8d" $s]
}

proc Now {} {
    # current date under default format
    global DateFormat
    
    return [Today $DateFormat]
}

set DAYSOF(1) 31 ; set DAYSOF(2) 28 ; set DAYSOF(3) 31 ; set DAYSOF(4) 30
set DAYSOF(5) 31 ; set DAYSOF(6) 30 ; set DAYSOF(7) 31 ; set DAYSOF(8) 31
set DAYSOF(9) 30 ; set DAYSOF(10) 31 ; set DAYSOF(11) 30 ; set DAYSOF(12) 31

proc DateToSecs {y m d h mn s} {
    # convert date to seconds ellapsed since beginning of $YEAR0 (a leap year)
    global DAYSOF YEAR0

    set days [expr 365*($y-$YEAR0)+$d-1]
    set yy $YEAR0
    while { $yy < $y } {
	if { $yy%100!=0 || $yy%400==0 } { incr days }
	incr yy 4
    }
    if { $m>2 && $y%4==0 && ($yy%100!=0 || $yy%400==0) } { incr days }
    incr m -1
    while { $m > 0 } { incr days $DAYSOF($m) ; incr m -1 }
    return [expr $days*86400+$h*3600+$mn*60+$s]
}

proc DateFromSecs {secs} {
    # build date from seconds ellapsed since beginning of $YEAR0 (a leap year)
    #  using default format
    global DateFormat

    return [DateFromSecsFmt $secs $DateFormat]
}

proc DateFromSecsFmt {secs fmt} {
    # build date from seconds ellapsed since beginning of $YEAR0 (a leap year)
    #  using given format
    global DAYSOF YEAR0

    set s [expr $secs%60] ; set x [expr ($secs-$s)/60]
    set mn [expr $x%60] ; set x [expr ($x-$mn)/60]
    set h [expr $x%24] ; set x [expr ($x-$h)/24]
    set y [expr int($x/365)] ; set yd [expr $y*365]
    set yy $YEAR0 ; incr y $yy
    while { $yy < $y } {
	if { $yy%100!=0 || $yy%400==0 } { incr yd }
	incr yy 4
    }
    if { $yd > $x } {
	incr y -1
	if { $y%4==0 && ($y%100!=0 || $y%400==0) } { incr yd -1 }
	incr yd -365
    }
    incr x [expr 1-$yd] ; set m 1
    while { $x > $DAYSOF($m) } {
	if { $m==2 && $y%4==0 && ($y%100!=0 || $y%400==0) } {
	    if { $x == 29 } { break }
	    incr x -1
	}
	incr x -$DAYSOF($m) ; incr m
    }
    return [FormatDate $fmt $y $m $x $h $mn $s]
}

proc Month {name} {
    # from month name to 0 (error) or 1..12
    global ALLMONTH

    for {set m 1} { $m<13 } { incr m } {
	if { [lsearch $ALLMONTH($m) $name] != -1 } { return $m }
    }
    return 0
}

proc UTCToLocalTime {y m d h mn s} {
    # convert from UTC to local time
    # return list with same elements as in the args
    global TimeOffset DAYSOF

    incr h $TimeOffset
    if { $h < 0 } {
	incr h 24
	if { [incr d -1] == 0 } {
	    if { [incr m -1] == 0 } {
		set m 12 ; set d 31 ; incr y -1
	    } else {
		if { $m == 2 && $y%4 == 0 && \
			($y%100 != 0 || $y%400 == 0) } {
		    set d 29
		} else {
		    set d $DAYSOF($m)
		}
	    }
	}
    } elseif { $h > 24 } {
	incr h -24
	if { [incr d] > $DAYSOF($m) } {
	    if { $m == 2 } {
		if { $y%4 != 0 || ($y%100 == 0 && $y%400 != 0) } {
		    set m 3 ; set d 1
		}
	    } elseif { $m == 12 } {
		set m 1 ; set d 1 ; incr y
	    } else {
		incr m ; set d 1
	    }
	}
    }
    return [list $y $m $d $h $mn $s]
}

proc CompatibleDates {y0 m0 d0 secs1} {
    # check whether two dates are compatible
    #  $y0, $m0, $d0 give one date (year, month name and day) and may
    #  be * (to match anything)
    #  $secs1 is the other date in seconds since $YEAR0
    # return 1 if compatible

    set dt1 [lindex [DateFromSecsFmt $secs1 DDMMMYYYY] 0]
    if { [string compare $d0 *] && $d0<10 } { set d0 0$d0 }
    if { [string match $d0-$m0-$y0 $dt1] } { return 1 }
    return 0
}
