#
#  gpsman --- GPS Manager: a manager for GPS receiver data
#
#  Copyright (c) 1999 Miguel Filgueiras (mig@ncc.up.pt) / Universidade do Porto
#
#    This program is free software; you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation; either version 2 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.
#
#  File: mapproj.tcl
#  Last change:  12 December 2000
#

## some information used here adopted from
#  PROJ4.0 - cartographic projection software
#      by Gerald I. Evenden (gie@kai.er.usgs.gov)
##
## information on the German Grid (Gauss-Krueger-Koordinatensystem) projection
# provided by Andreas Lange (Andreas.Lange@Rhein-Main.de)
##

## "projection" here means the conversion from geodetic position to
# planar Cartesian coordinates in the terrain; these coordinates are
# then converted to map coordinates by the transformations defined in
# maptransf.tcl
#
#  No changes in $Datum will occur after initialization
##

## when adding new projection procedures these variables must be updated
# as well as MAPKNOWNPROJS set in main.tcl, and PRJ$name in files lang*.tcl;
# see also the comments preceding each procedure for the UTM projection
##

     # indices of MPData array used for each main projection
set MAPPROJDATA(UTM) {UTMzone long0}
set MAPPROJDATA(TM) {lat0 long0 k0}
set MAPPROJDATA(GKK) {gkklong0}

     # new types here imply changing proc BadProjParam below
array set MAPPROJDTYPE {
    lat0      range=-90,90
    long0     range=-180,180
    k0        float>0
    gkklong0  list=0,3,6,9,12,15
}

     # particular cases of main projections:
     #  CMP - projection used in Portuguese Military Maps
     #  BNG - projection used for the British National Grid
     #  BWI - projection used for the British West Indies Grid
     #  ITM - projection used for the Irish TM Grid
array set MAPPARTPROJ {
    CMP TM
    BNG TM
    BWI TM
    ITM TM
}

     # parameters aligned with MAPPROJDATA above
array set MAPPARTPDATA {
    CMP {39.66666666666667 -8.13190611111111 1}
    BNG {49.0 -2.0 0.9996012717}
    BWI {0 -62.0 0.9995}
    ITM {53.5 -8.0 1.000035}
}

## initialization of particular cases of main projections

proc MapInitPartProj {ps} {
    # set fixed parameters for a particular case of a main projection
    # return projection of first position
    global MapProjection MPData MAPPARTPROJ MAPPARTPDATA MAPPROJDATA

    set mp $MAPPARTPROJ($MapProjection)
    foreach e $MAPPROJDATA($mp) v $MAPPARTPDATA($MapProjection) {
	set MPData($e) $v
    }
    set p [lindex $ps 0]
    set latd [lindex $p 0] ; set longd [lindex $p 1] ; set datum [lindex $p 2]
    return [MP${mp}Point $latd $longd $datum]
}

## setting up use of a projection

proc MapProjectionIs {proj} {
    # set global variables so that map projection is $proj
    #  $proj in $MAPKNOWNPROJS
    global MapProjection MapProjTitle MapProjInitProc MAPPARTPROJ TXT

    set MapProjection $proj ; set MapProjTitle $TXT(PRJ$proj)
    # names of projection procs
    if { [catch "set mp $MAPPARTPROJ($proj)"] } {
	# main projection
	set MapProjInitProc MP${proj}Init
    } else {
	# particular case of a main projection
	set MapProjInitProc MapInitPartProj
	set proj $mp
    }
    foreach a "Point Invert InvUTM" {
	global MapProj${a}Proc

	set MapProj${a}Proc MP${proj}$a
    }
    return
}

## changing parameters

proc BadProjParam {prm value} {
    # check value for projection parameter
    global MAPPROJDTYPE MESS TXT

    if { [regexp {^ *-?[0-9]+(\.[0-9]+)? *$} $value] } {
	switch -glob $MAPPROJDTYPE($prm) {
	    range=*,* {
		regexp {range=([^,]+),(.*)} $MAPPROJDTYPE($prm) z mn mx
		set e [expr $value<$mn || $value>$mx]
	    }
	    float>* {
		regsub float> $MAPPROJDTYPE($prm) "" mn
		set e [expr $value<$mn]
	    }
	    list=* {
		regsub list= $MAPPROJDTYPE($prm) "" l
		set e 1
		foreach v [split $l ,] {
		    if { $value == $v } { set e 0 ; break }
		}
	    }
	}
    } else { set e 1 }
    if { $e } {
	GMMessage [format $MESS(badprojprm) $TXT($prm)]
	return 1
    }
    return 0
}

proc ProjParamsChanged {proj} {
    # open dialog for confirming the projection parameters
    #  $proj in $MAPKNOWNPROJS
    # may turn ASKPROJPARAMS to false in current session
    # return 1 if there were changes
    global MPData MAPPROJDATA MAPPROJDTYPE TXT MESS COLOUR ASKPROJPARAMS \
	    EPOSX EPOSY

    set waitp 1
    toplevel .proj
    wm protocol .proj WM_DELETE_WINDOW "set waitp 0"
    wm title .proj "GPS Manager: $TXT(projection)"
    wm transient .proj
    wm geometry .proj +$EPOSX+$EPOSY
    bind .proj <Key-Return> "set waitp 0"

    frame .proj.fr -relief flat -borderwidth 5 -bg $COLOUR(selbg)
    label .proj.fr.title -text "???" -relief sunken
    label .proj.fr.text -text $MESS(projchg)
    frame .proj.fr.frsel -relief flat -borderwidth 0
    set n 0
    foreach p $MAPPROJDATA($proj) {
	if { [string compare $MAPPROJDTYPE($p) reserved] } {
	    label .proj.fr.frsel.l$p -text $TXT($p)
	    entry .proj.fr.frsel.e$p
	    .proj.fr.frsel.e$p insert 0 $MPData($p)
	    grid config .proj.fr.frsel.l$p -column 0 -row $n -sticky w
	    grid config .proj.fr.frsel.e$p -column 1 -row $n -sticky snew
	    incr n
	}
    }
    if { $n == 0 } {
	destroy .proj
	return 0
    }
    checkbutton .proj.fr.stop -text $TXT(dontaskagain) \
	    -variable ASKPROJPARAMS -onvalue 0 -offvalue 1 \
	    -selectcolor $COLOUR(check)
    frame .proj.fr.frbs -relief flat -borderwidth 0
    button .proj.fr.frbs.ok -text Ok -command "set waitp 0"
    button .proj.fr.frbs.rv -text $TXT(revert) \
	    -command "ProjParamsRevert $proj"
    pack .proj.fr -side top
    pack .proj.fr.frbs.ok .proj.fr.frbs.rv -side left -padx 5
    pack .proj.fr.title .proj.fr.text .proj.fr.frsel .proj.fr.stop \
	    .proj.fr.frbs -side top -pady 5
    update idletasks
    grab .proj
    while 1 {
	tkwait variable waitp
	set l "" ; set r 1
	foreach p $MAPPROJDATA($proj) {
	    if { [string compare [set t $MAPPROJDTYPE($p)] reserved] } {
		set nv [.proj.fr.frsel.e$p get]
		if { $nv != $MPData($p) } {
		    if { [BadProjParam $p $nv] } {
			set r error
			break
		    }
		    lappend l [list $p $nv]
		    set r 0
		}
	    }
	}
	switch $r {
	    1 {
		foreach pv $l {
		    set MPData([lindex $pv 0]) [lindex $pv 1]
		}
	    }
	    error { continue }
	}
	break
    }
    destroy .proj
    update idletasks
    return $r
}

proc ProjParamsRevert {proj} {
    # revert to initial values in dialog on the projection parameters
    global MPData MAPPROJDATA MAPPROJDTYPE

    foreach p $MAPPROJDATA($proj) {
	if { [string compare $MAPPROJDTYPE($p) reserved] } {
	    .proj.fr.frsel.e$p delete 0 end
	    .proj.fr.frsel.e$p insert 0 $MPData($p)
	}
    }
    return
}

## map projection procedures for UTM projection

# initialization projection proc sets initial values for the projection
#  proc given a list of latd,longd,datum to be projected
# will be called only when the map is empty
# should call proc ProjParamsChanged if its parameters may be changed by
#  the user and if $ASKPROJPARAMS is true (see MPTMInit)
# must return projection of first position

proc MPUTMInit {ps} {
    # set parameters for UTM projection
    # only the first position is used
    # no parameters can be changed by the user
    # return projection of first position
    global Datum MPData

    set p [lindex $ps 0]
    set latd [lindex $p 0] ; set longd [lindex $p 1] ; set datum [lindex $p 2]
    if { [string compare $datum $Datum] } {
	set p [ConvertDatum $latd $longd $datum $Datum DDD]
	set latd [lindex $p 0] ; set longd [lindex $p 1]
    }
    set p [DegreesToUTM $latd $longd $Datum]
    set ze [lindex $p 0] ; set zn [lindex $p 1]
    set MPData(UTMzone) [format %d%s $ze $zn]
    set MPData(long0) [expr 6*int(floor($longd/6.0))+3]
    return [list [lindex $p 2] [lindex $p 3]]
}

# projection proc is a function of latd,longd,datum to planar Cartesian
#  coordinates x,y in the terrain for $Datum

proc MPUTMPoint {latd longd datum} {
    # compute planar Cartesian coordinates under $Datum for given position
    global Datum MPData

    if { [string compare $datum $Datum] } {
	set p [ConvertDatum $latd $longd $datum $Datum DDD]
	set latd [lindex $p 0] ; set longd [lindex $p 1]
    }
    set p [DegreesToUTM $latd $longd $Datum]
    set pze [lindex $p 0] ; set pzn [lindex $p 1]
    set px [lindex $p 2] ; set py [lindex $p 3]
    scan $MPData(UTMzone) %d%s mze mzn
    if { $mze!=$pze || [string compare $mzn $pzn] } {
	set cs [CompUTMOnZone $latd $longd $MPData(long0)]
	set px [lindex $cs 0] ; set py [lindex $cs 1]
    }
    return [list $px $py]
}

# invert projection proc is a function of planar Cartesian coordinates in the
#  terrain for $Datum to position in given format also for $Datum

proc MPUTMInvert {x y pfmt} {
    # return position in $pfmt format for point at $x,$y in the terrain
    #  for $Datum
    global Datum MPData

    scan $MPData(UTMzone) %d%s ze zn
    set p [UTMToDegrees $ze $zn $x $y $Datum]
    set p [CreatePos [lindex $p 0] [lindex $p 1] $pfmt [PosType $pfmt] $Datum]
    return $p
}

# convert to UTM projection proc is a function of planar Cartesian coordinates
#  in the terrain for $Datum to UTM position also for $Datum

proc MPUTMInvUTM {x y} {
    # return UTM position for point at $x,$y in the terrain, for $Datum
    global Datum MPData

    scan $MPData(UTMzone) %d%s pze pzn
    set p [UTMToDegrees $pze $pzn $x $y $Datum]
    return [DegreesToUTM [lindex $p 0] [lindex $p 1] $Datum]
}

## Transverse Mercator projection

proc MPTMInit {ps} {
    # set parameters for TM projection
    # take averages of latitudes and of longitudes as central coordinates
    #  and use k0=0.9996
    # return projection of first position
    global Datum MPData ASKPROJPARAMS

    set p [lindex $ps 0]
    set latd [lindex $p 0] ; set longd [lindex $p 1] ; set datum [lindex $p 2]
    if { [string compare $datum $Datum] } {
	set p [ConvertDatum $latd $longd $datum $Datum DDD]
	set latd [lindex $p 0] ; set longd [lindex $p 1]
    }
    set sla $latd ; set slo $longd ; set n 1
    foreach p [lrange $ps 1 end] {
	set la [lindex $p 0] ; set lo [lindex $p 1] ; set datum [lindex $p 2]
	if { [string compare $datum $Datum] } {
	    set p [ConvertDatum $la $lo $datum $Datum DDD]
	    set la [lindex $p 0] ; set lo [lindex $p 1]
	}
	set sla [expr $sla+$la] ; set slo [expr $slo+$lo]
	incr n
    }
    set lat0 [expr 1.0*$sla/$n] ; set long0 [expr 1.0*$slo/$n]
    set MPData(lat0) $lat0 ; set MPData(long0) $long0 ; set MPData(k0) 0.9996
    if { $ASKPROJPARAMS } {
	ProjParamsChanged TM
    }
    return [ConvToTM $latd $longd \
	             $MPData(lat0) $MPData(long0) $MPData(k0) $Datum]
}

proc MPTMPoint {latd longd datum} {
    # compute planar Cartesian coordinates under $Datum for given position
    global Datum MPData

    if { [string compare $datum $Datum] } {
	set p [ConvertDatum $latd $longd $datum $Datum DDD]
	set latd [lindex $p 0] ; set longd [lindex $p 1]
    }
    return [ConvToTM $latd $longd \
	             $MPData(lat0) $MPData(long0) $MPData(k0) $Datum]
}

proc MPTMInvert {x y pfmt} {
    # return position in $pfmt format for point at $x,$y in the terrain
    #  for $Datum
    global Datum MPData

    set p [ConvFromTM $x $y $MPData(lat0) $MPData(long0) $MPData(k0) $Datum]
    set p [CreatePos [lindex $p 0] [lindex $p 1] $pfmt [PosType $pfmt] $Datum]
    return $p
}

proc MPTMInvUTM {x y} {
    # return UTM position for point at $x,$y in the terrain, for $Datum
    global Datum MPData

    set p [ConvFromTM $x $y $MPData(lat0) $MPData(long0) $MPData(k0) $Datum]
    return [DegreesToUTM [lindex $p 0] [lindex $p 1] $Datum]
}

## German Grid (Gauss-Krueger-Koordinatensystem) projection
# Information provided by Andreas Lange (Andreas.Lange@Rhein-Main.de)
# This is a Transverse Mercator projection with lat0=0, lon0=zone*3
#  (0, 3, 6, 9, 12, 15E), a false easting of zone*1000000 + 500000 and scale
#  factor at central meridian of k0=1.0

proc MPGKKInit {ps} {
    # set parameters for GKK projection
    # take averages of longitudes to find central longitude
    #  and use lat0=0, k0=1
    # return projection of first position
    global Datum MPData ASKPROJPARAMS

    set p [lindex $ps 0]
    set latd [lindex $p 0] ; set longd [lindex $p 1] ; set datum [lindex $p 2]
    if { [string compare $datum $Datum] } {
	set p [ConvertDatum $latd $longd $datum $Datum DDD]
	set latd [lindex $p 0] ; set longd [lindex $p 1]
    }
    set slo $longd ; set n 1
    foreach p [lrange $ps 1 end] {
	set lo [lindex $p 1] ; set datum [lindex $p 2]
	if { [string compare $datum $Datum] } {
	    set p [ConvertDatum [lindex $p 0] $lo $datum $Datum DDD]
	    set lo [lindex $p 1]
	}
	set slo [expr $slo+$lo]
	incr n
    }
    set lat0 0
    if { [set long0 [expr 1.0*$slo/$n]] < 0 } {
	set long0 0
    } elseif { $long0 > 15 } {
	set long0 15
    } else {
	set long0 [expr int(($long0+1.5)/3)*3]
    }
    set MPData(gkklong0) $long0
    if { $ASKPROJPARAMS } {
	ProjParamsChanged GKK
    }
    return [ConvToTM $latd $longd 0 $long0 1 $Datum]
}

proc MPGKKPoint {latd longd datum} {
    # compute planar Cartesian coordinates under $Datum for given position
    global Datum MPData

    if { [string compare $datum $Datum] } {
	set p [ConvertDatum $latd $longd $datum $Datum DDD]
	set latd [lindex $p 0] ; set longd [lindex $p 1]
    }
    return [ConvToTM $latd $longd 0 $MPData(gkklong0) 1 $Datum]
}

proc MPGKKInvert {x y pfmt} {
    # return position in $pfmt format for point at $x,$y in the terrain
    #  for $Datum
    global Datum MPData

    set p [ConvFromTM $x $y 0 $MPData(gkklong0) 1 $Datum]
    set p [CreatePos [lindex $p 0] [lindex $p 1] $pfmt [PosType $pfmt] $Datum]
    return $p
}

proc MPGKKInvUTM {x y} {
    # return UTM position for point at $x,$y in the terrain, for $Datum
    global Datum MPData

    set p [ConvFromTM $x $y 0 $MPData(gkklong0) 1 $Datum]
    return [DegreesToUTM [lindex $p 0] [lindex $p 1] $Datum]
}

